(ns coderloop.top-five)


(use 'rlm.shell-inspect)
(use 'clojure.contrib.profile)
(use '[clojure.contrib [duck-streams :only [file-str read-lines *default-encoding*]]])
(import '(java.nio ByteBuffer CharBuffer)
	'(java.io PushbackReader InputStream InputStreamReader
		  FileInputStream))



;;  ^{:doc "Name of the default encoding to use when reading & writing.
;;   Default is UTF-8."
;;     :tag "java.lang.String"}
;;  *default-encoding* "US-ASCII")
;;(set! clojure.contrib.duck-streams/*default-encoding* "US-ASCII")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

 (def a (file-str "/home/r/coderloop-test/topfive-a.in"))
 (def b (file-str "/home/r/coderloop-test/topfive-b.in"))
 (def c (file-str "/home/r/coderloop-test/topfive-c.in"))
 (def d (file-str "/home/r/coderloop-test/topfive-d.in"))
 (def e (file-str "/home/r/coderloop-test/topfive-e.in"))
 (def f (file-str "/home/r/coderloop-test/topfive-f.in"))

(defn get-query-slow [s]
     (nth (re-matches #"^.*, query=(.*)]$" s) 1))

(defn #^java.lang.String get-query [#^java.lang.String s]
   (.substring s (clojure.core/+ (.lastIndexOf s "query=") 6) (clojure.core/- (.length s) 1)))


(defn chuncked-pmap
  "helps with paralleization of functions that don't take
   very much time to execute"  
  [n f coll]
  (apply concat
   (pmap
    (fn [coll]
      (doall (map f coll)))
    (partition-all n coll))))


;; (def *main-map* (atom {}))

;; (declare parse-lines tally!)

;; (defn parse-lines [lines]
;;   (map get-query lines))

;; (defn tally! [state parsed-lines]
;;   (doseq [element parsed-lines]
;;     (bump! state element 1)))


;; (defn analyze-log [file]
;;   (let [chunk-count (int (/ (.length #^java.io.File file) (* 32 1024 1024)))
;; 	state (atom {})]
;;     (dorun
;;      (pmap (fn [[idx [start end]]]
;; 	     (println (str "Chunk " idx "/" chunk-count
;; 			   " (" start " -> " end ")"))
;; 	     (->> (read-lines-range file start end)
;; 		  (parse-lines)
;; 		  (tally! state)))
;; 	   (indexed (chunk-file file chunk-count))))
;;     state))




;; (defn preduce-full
;;   "a parallel reduce. Because it is parallel, the reducing
;;    function must make sense in a tree-reducing context"
;;   [num-chunks mapping-function f coll]
;;   (loop [prev coll]
;;     (let [new-coll
;; 	  (concat
;; 	   (mapping-function 
;; 	    (partial apply f)
;; 	    (partition-all num-chunks prev)))]
;;       (if (not (next new-coll))
;; 	(first new-coll)
;; 	(recur new-coll)))))





;;; correctness before speed:
(defn top-five-0 [file]
  (binding [*default-encoding* "US-ASCII"]
    
    (map
     first
     (take
      5
      (reverse
       (sort-by second
		(reduce
		 (partial merge-with +)
		 (map frequencies
		      (partition-all 2 (map get-query (read-lines file)))))))))))
"Elapsed time: 34440.153889 msecs"
"Elapsed time: 21072.141885 msecs"
"Elapsed time: 21041.711202 msecs"
;; now let's do speed:
(defn top-five-1 [file]
  (map
   first
   (take
    5
    (sort-by (comp - second)
	     (reduce
	      (partial merge-with +)
	      (map frequencies
		   (partition-all
		    800 (map get-query (read-lines file)))))))))


(defn top-five-3 [file]
  (map
   first
   (take 5
	 (sort-by (comp - second)
		  (frequencies
		   (map get-query (read-lines file)))))))



"Elapsed time: 12877.194194 msecs"
"Elapsed time: 12282.975191 msecs"
;; got a factor of 2x increase in speed from a larger partition and chuncked-pmap.
;; still too slow....

;;; let's try the "make it all one big function" approach
;;(set! *warn-on-reflection* true)

(defn top-five-2 [file]

  (let [*top-5* (atom (vector :a :b :c :d :e))
	*data* (atom {})]
    
    (letfn [(top-ten-helper
	     [line]
	     (let [query (get-query line)]
	       (swap! *data* #(assoc % query (inc (get @*data* query 0))))
	       (if (> (@*data* query 0) (get @*data* (nth @*top-5* 0) 0))
		 (do

		   (if-not (contains? (set @*top-5*) query)
		     (do 
		       (swap! *top-5* #(assoc % 0 query))
		       (swap! *top-5*
			      (fn [v] (vec (sort-by #(get @*data* % 0) v))))))))))]

      (dorun (chuncked-pmap 800 top-ten-helper (read-lines file)))
      (swap! *top-5*
	     (fn [v] (vec (sort-by #(get @*data* % 0) v))))
      (reverse @*top-5*))))
"Elapsed time: 10735.897831 msecs" ;; with chuncked-pmap

	
    
  
(if (command-line?)
  (do
    (dorun (map println (top-five-3 (file-str (first *command-line-args*)))))
    (System/exit 0)))
